@@ -1,5 +1,5 @@
 /*
-
+Copyright 2021.
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -19,6 +19,7 @@ package controllers
 import (
 	"context"
 	"fmt"
+	"os"
 	"strconv"
 	"time"
 
@@ -38,15 +39,18 @@ import (
 
 	operatorv1alpha1 "github.com/gatekeeper/gatekeeper-operator/api/v1alpha1"
 	"github.com/gatekeeper/gatekeeper-operator/controllers/merge"
+	"github.com/gatekeeper/gatekeeper-operator/pkg/platform"
 	"github.com/gatekeeper/gatekeeper-operator/pkg/util"
 )
 
 const (
 	defaultGatekeeperCrName           = "gatekeeper"
 	openshiftAssetsDir                = "openshift/"
+	GatekeeperImageEnvVar             = "RELATED_IMAGE_GATEKEEPER"
 	NamespaceFile                     = "v1_namespace_gatekeeper-system.yaml"
-	AssignCRDFile                     = "apiextensions.k8s.io_v1beta1_customresourcedefinition_assign.mutations.gatekeeper.sh.yaml"
-	AssignMetadataCRDFile             = "apiextensions.k8s.io_v1beta1_customresourcedefinition_assignmetadata.mutations.gatekeeper.sh.yaml"
+	AssignCRDFile                     = "apiextensions.k8s.io_v1_customresourcedefinition_assign.mutations.gatekeeper.sh.yaml"
+	AssignMetadataCRDFile             = "apiextensions.k8s.io_v1_customresourcedefinition_assignmetadata.mutations.gatekeeper.sh.yaml"
+	MutatorPodStatusCRDFile           = "apiextensions.k8s.io_v1_customresourcedefinition_mutatorpodstatuses.status.gatekeeper.sh.yaml"
 	AuditFile                         = "apps_v1_deployment_gatekeeper-audit.yaml"
 	WebhookFile                       = "apps_v1_deployment_gatekeeper-controller-manager.yaml"
 	ClusterRoleFile                   = "rbac.authorization.k8s.io_v1_clusterrole_gatekeeper-manager-role.yaml"
@@ -54,11 +58,13 @@ const (
 	RoleFile                          = "rbac.authorization.k8s.io_v1_role_gatekeeper-manager-role.yaml"
 	RoleBindingFile                   = "rbac.authorization.k8s.io_v1_rolebinding_gatekeeper-manager-rolebinding.yaml"
 	ServerCertFile                    = "v1_secret_gatekeeper-webhook-server-cert.yaml"
-	ValidatingWebhookConfiguration    = "admissionregistration.k8s.io_v1beta1_validatingwebhookconfiguration_gatekeeper-validating-webhook-configuration.yaml"
-	MutatingWebhookConfiguration      = "admissionregistration.k8s.io_v1beta1_mutatingwebhookconfiguration_gatekeeper-mutating-webhook-configuration.yaml"
+	ValidatingWebhookConfiguration    = "admissionregistration.k8s.io_v1_validatingwebhookconfiguration_gatekeeper-validating-webhook-configuration.yaml"
+	MutatingWebhookConfiguration      = "admissionregistration.k8s.io_v1_mutatingwebhookconfiguration_gatekeeper-mutating-webhook-configuration.yaml"
 	ValidationGatekeeperWebhook       = "validation.gatekeeper.sh"
 	CheckIgnoreLabelGatekeeperWebhook = "check-ignore-label.gatekeeper.sh"
 	MutationGatekeeperWebhook         = "mutation.gatekeeper.sh"
+	AuditDeploymentName               = "gatekeeper-audit"
+	WebhookDeploymentName             = "gatekeeper-controller-manager"
 	managerContainer                  = "manager"
 	LogLevelArg                       = "--log-level"
 	AuditIntervalArg                  = "--audit-interval"
@@ -69,20 +75,26 @@ const (
 	EmitAdmissionEventsArg            = "--emit-admission-events"
 	ExemptNamespaceArg                = "--exempt-namespace"
 	EnableMutationArg                 = "--enable-mutation"
+	OperationArg                      = "--operation"
+	OperationMutationStatus           = "mutation-status"
+	DisabledBuiltinArg                = "--disable-opa-builtin"
 )
 
 var (
 	orderedStaticAssets = []string{
 		NamespaceFile,
-		"apiextensions.k8s.io_v1beta1_customresourcedefinition_configs.config.gatekeeper.sh.yaml",
-		"apiextensions.k8s.io_v1beta1_customresourcedefinition_constrainttemplates.templates.gatekeeper.sh.yaml",
-		"apiextensions.k8s.io_v1beta1_customresourcedefinition_constrainttemplatepodstatuses.status.gatekeeper.sh.yaml",
-		"apiextensions.k8s.io_v1beta1_customresourcedefinition_constraintpodstatuses.status.gatekeeper.sh.yaml",
+		"v1_resourcequota_gatekeeper-critical-pods.yaml",
+		"apiextensions.k8s.io_v1_customresourcedefinition_configs.config.gatekeeper.sh.yaml",
+		"apiextensions.k8s.io_v1_customresourcedefinition_constrainttemplates.templates.gatekeeper.sh.yaml",
+		"apiextensions.k8s.io_v1_customresourcedefinition_constrainttemplatepodstatuses.status.gatekeeper.sh.yaml",
+		"apiextensions.k8s.io_v1_customresourcedefinition_constraintpodstatuses.status.gatekeeper.sh.yaml",
 		AssignCRDFile,
 		AssignMetadataCRDFile,
+		MutatorPodStatusCRDFile,
 		ServerCertFile,
 		"v1_serviceaccount_gatekeeper-admin.yaml",
 		"policy_v1beta1_podsecuritypolicy_gatekeeper-admin.yaml",
+		"policy_v1beta1_poddisruptionbudget_gatekeeper-controller-manager.yaml",
 		ClusterRoleFile,
 		ClusterRoleBindingFile,
 		RoleFile,
@@ -96,9 +108,10 @@ var (
 		MutatingWebhookConfiguration,
 	}
 
-	mutatingCRDs = []string{
+	MutatingCRDs = []string{
 		AssignCRDFile,
 		AssignMetadataCRDFile,
+		MutatorPodStatusCRDFile,
 	}
 )
 
@@ -108,7 +121,7 @@ type GatekeeperReconciler struct {
 	Log          logr.Logger
 	Scheme       *runtime.Scheme
 	Namespace    string
-	PlatformName util.PlatformType
+	PlatformInfo platform.PlatformInfo
 }
 
 type crudOperation uint32
@@ -118,7 +131,7 @@ const (
 	delete crudOperation = iota
 )
 
-// Gatekeeper Operator RBAC permissions to manager Gatekeeper custom resource
+// Gatekeeper Operator RBAC permissions to manage Gatekeeper custom resource
 // +kubebuilder:rbac:groups=operator.gatekeeper.sh,resources=gatekeepers,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=operator.gatekeeper.sh,resources=gatekeepers/status,verbs=get;update;patch
 // +kubebuilder:rbac:groups=operator.gatekeeper.sh,resources=gatekeepers/finalizers,verbs=delete;get;update;patch
@@ -129,6 +142,7 @@ const (
 
 // Cluster Scoped
 // +kubebuilder:rbac:groups=*,resources=*,verbs=get;list;watch
+// +kubebuilder:rbac:groups=core,resources=namespaces,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=apiextensions.k8s.io,resources=customresourcedefinitions,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=config.gatekeeper.sh,resources=configs,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=config.gatekeeper.sh,resources=configs/status,verbs=get;update;patch
@@ -144,12 +158,21 @@ const (
 // +kubebuilder:rbac:groups=admissionregistration.k8s.io,resources=mutatingwebhookconfigurations,verbs=get;list;watch;create;update;patch;delete
 
 // Namespace Scoped
-// +kubebuilder:rbac:groups=core,namespace="system",resources=secrets;serviceaccounts;services,verbs=get;list;watch;create;update;patch;delete
+// +kubebuilder:rbac:groups=core,namespace="system",resources=secrets;serviceaccounts;services;resourcequotas,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=rbac.authorization.k8s.io,namespace="system",resources=roles;rolebindings,verbs=get;list;watch;create;update;patch;delete
 // +kubebuilder:rbac:groups=apps,namespace="system",resources=deployments,verbs=get;list;watch;create;update;patch;delete
+// +kubebuilder:rbac:groups=policy,namespace="system",resources=poddisruptionbudgets,verbs=create;delete;update;use
 
-func (r *GatekeeperReconciler) Reconcile(req ctrl.Request) (ctrl.Result, error) {
-	ctx := context.Background()
+// Reconcile is part of the main kubernetes reconciliation loop which aims to
+// move the current state of the cluster closer to the desired state.
+// TODO(user): Modify the Reconcile function to compare the state specified by
+// the Gatekeeper object against the actual cluster state, and then
+// perform operations to make the cluster state reflect the state specified by
+// the user.
+//
+// For more details, check Reconcile and its Result here:
+// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.9.2/pkg/reconcile
+func (r *GatekeeperReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
 	logger := r.Log.WithValues("gatekeeper", req.NamespacedName)
 	logger.Info("Reconciling Gatekeeper")
 
@@ -164,13 +187,16 @@ func (r *GatekeeperReconciler) Reconcile(req ctrl.Request) (ctrl.Result, error)
 	err := r.Get(ctx, req.NamespacedName, gatekeeper)
 	if err != nil {
 		if apierrors.IsNotFound(err) {
-
 			return ctrl.Result{}, nil
 		}
-
 		return ctrl.Result{}, err
 	}
 
+	if gatekeeper.Spec.Image != nil && gatekeeper.Spec.Image.Image != nil {
+		logger.Info("WARNING: operator.gatekeeper.sh/v1alpha1 Gatekeeper spec.image.image field is no longer supported and will be removed in a future release.",
+			"spec.image.image", gatekeeper.Spec.Image.Image)
+	}
+
 	err, requeue := r.deployGatekeeperResources(gatekeeper)
 	if err != nil {
 		return ctrl.Result{}, errors.Wrap(err, "Unable to deploy Gatekeeper resources")
@@ -181,13 +207,14 @@ func (r *GatekeeperReconciler) Reconcile(req ctrl.Request) (ctrl.Result, error)
 	return ctrl.Result{}, nil
 }
 
+// SetupWithManager sets up the controller with the Manager.
 func (r *GatekeeperReconciler) SetupWithManager(mgr ctrl.Manager) error {
 	return ctrl.NewControllerManagedBy(mgr).
 		For(&operatorv1alpha1.Gatekeeper{}).
 		WithEventFilter(predicate.Funcs{
 			UpdateFunc: func(e event.UpdateEvent) bool {
-				oldGeneration := e.MetaOld.GetGeneration()
-				newGeneration := e.MetaNew.GetGeneration()
+				oldGeneration := e.ObjectOld.GetGeneration()
+				newGeneration := e.ObjectNew.GetGeneration()
 
 				return oldGeneration != newGeneration
 			},
@@ -200,48 +227,61 @@ func (r *GatekeeperReconciler) SetupWithManager(mgr ctrl.Manager) error {
 }
 
 func (r *GatekeeperReconciler) deployGatekeeperResources(gatekeeper *operatorv1alpha1.Gatekeeper) (error, bool) {
-	deleteAssets, applyAssets, webhookAssets := getStaticAssets(gatekeeper)
+	deleteWebhookAssets, applyOrderedAssets, applyWebhookAssets, deleteCRDAssets := getStaticAssets(gatekeeper)
 
-	for _, d := range deleteAssets {
-		obj, err := util.GetManifestObject(d)
-		if err != nil {
-			return err, false
-		}
-
-		if err = r.crudResource(obj, gatekeeper, delete); err != nil {
-			return err, false
-		}
+	if err := r.deleteAssets(deleteWebhookAssets, gatekeeper); err != nil {
+		return err, false
 	}
-	// Checking for deployment before deploying assets to avoid cert rotator errors
+
+	// Checking for deployment before deploying assets or deleting CRDs to
+	// avoid transient errors e.g. cert rotator errors, removing required CRD
+	// resources, etc.
 	err, requeue := r.validateWebhookDeployment()
 	if err != nil {
 		return err, false
 	}
-	for _, asset := range applyAssets {
-		err := r.applyAsset(gatekeeper, asset, false)
-		if err != nil {
-			return err, false
-		}
+
+	if err := r.applyAssets(applyOrderedAssets, gatekeeper, false); err != nil {
+		return err, false
 	}
 
-	for _, asset := range webhookAssets {
-		err := r.applyAsset(gatekeeper, asset, requeue)
-		if err != nil {
-			return err, false
-		}
+	if err := r.applyAssets(applyWebhookAssets, gatekeeper, requeue); err != nil {
+		return err, false
 	}
+
+	if err := r.deleteAssets(deleteCRDAssets, gatekeeper); err != nil {
+		return err, false
+	}
+
 	return nil, requeue
 }
 
+func (r *GatekeeperReconciler) deleteAssets(assets []string, gatekeeper *operatorv1alpha1.Gatekeeper) error {
+	for _, a := range assets {
+		obj, err := util.GetManifestObject(a)
+		if err != nil {
+			return err
+		}
+
+		if err = r.crudResource(obj, gatekeeper, delete); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func (r *GatekeeperReconciler) applyAssets(assets []string, gatekeeper *operatorv1alpha1.Gatekeeper, controllerDeploymentPending bool) error {
+	for _, a := range assets {
+		err := r.applyAsset(gatekeeper, a, controllerDeploymentPending)
+		if err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
 func (r *GatekeeperReconciler) applyAsset(gatekeeper *operatorv1alpha1.Gatekeeper, asset string, controllerDeploymentPending bool) error {
-	// Handle special cases in switch below.
-	switch {
-	case asset == NamespaceFile && !r.isOpenShift():
-		// Ignore the namespace resource on Kubernetes as we default to use
-		// the same namespace as the operator, which by definition is
-		// already created as a result of executing this code.
-		return nil
-	case asset == RoleFile && r.isOpenShift():
+	if asset == RoleFile && r.isOpenShift() {
 		asset = openshiftAssetsDir + asset
 	}
 
@@ -261,7 +301,7 @@ func (r *GatekeeperReconciler) applyAsset(gatekeeper *operatorv1alpha1.Gatekeepe
 }
 
 func (r *GatekeeperReconciler) validateWebhookDeployment() (error, bool) {
-	r.Log.Info(fmt.Sprintf("Validating %s deployment status", WebhookFile))
+	r.Log.Info(fmt.Sprintf("Validating %s deployment status", WebhookDeploymentName))
 
 	ctx := context.Background()
 	obj, err := util.GetManifestObject(WebhookFile)
@@ -279,7 +319,7 @@ func (r *GatekeeperReconciler) validateWebhookDeployment() (error, bool) {
 	err = r.Get(ctx, namespacedName, deployment)
 	if err != nil {
 		if apierrors.IsNotFound(err) {
-			r.Log.Info("Deployment not found, requeuing ...")
+			r.Log.Info("Deployment not found, will set webhook failure policy to ignore and requeue...")
 			return nil, true
 		}
 		return err, false
@@ -296,39 +336,42 @@ func (r *GatekeeperReconciler) validateWebhookDeployment() (error, bool) {
 		return err, false
 	}
 	if !ok {
-		return nil, true // State/readyReplicas might not yet be populated
+		r.Log.Info("Deployment status.readyReplicas not found or populated yet, will set webhook failure policy to ignore and requeue ...")
+		return nil, true
 	}
 	if replicas == readyReplicas {
 		r.Log.Info("Deployment validation successful, all replicas ready", "replicas", replicas, "readyReplicas", readyReplicas)
 		return nil, false
 	}
-	r.Log.Info("Deployment replicas not ready, requeuing ...", "replicas", replicas, "readyReplicas", readyReplicas)
+	r.Log.Info("Deployment replicas not ready, will set webhook failure policy to ignore and requeue ...",
+		"replicas", replicas, "readyReplicas", readyReplicas)
 	return nil, true
 }
 
-func getStaticAssets(gatekeeper *operatorv1alpha1.Gatekeeper) (deleteAssets, applyAssets, webhookAssets []string) {
+func getStaticAssets(gatekeeper *operatorv1alpha1.Gatekeeper) (deleteWebhookAssets, applyOrderedAssets, applyWebhookAssets, deleteCRDAssets []string) {
 	validatingWebhookEnabled := gatekeeper.Spec.ValidatingWebhook == nil || *gatekeeper.Spec.ValidatingWebhook == operatorv1alpha1.WebhookEnabled
 	mutatingWebhookEnabled := gatekeeper.Spec.MutatingWebhook != nil && mutatingWebhookEnabled(gatekeeper.Spec.MutatingWebhook)
-	deleteAssets = make([]string, 0)
-	applyAssets = make([]string, 0)
-	webhookAssets = make([]string, 0)
+	deleteWebhookAssets = make([]string, 0)
+	applyOrderedAssets = make([]string, 0)
+	applyWebhookAssets = make([]string, 0)
+	deleteCRDAssets = make([]string, 0)
 	// Copy over our set of ordered static assets so we maintain its
 	// immutability.
-	applyAssets = append(applyAssets, orderedStaticAssets...)
-	webhookAssets = append(webhookAssets, webhookStaticAssets...)
+	applyOrderedAssets = append(applyOrderedAssets, orderedStaticAssets...)
+	applyWebhookAssets = append(applyWebhookAssets, webhookStaticAssets...)
 
 	if !validatingWebhookEnabled {
-		// Remove ValidatingWebhookConfiguration resource
-		deleteAssets = append(deleteAssets, ValidatingWebhookConfiguration)
-		webhookAssets = getSubsetOfAssets(webhookAssets, ValidatingWebhookConfiguration)
+		// Remove and apply ValidatingWebhookConfiguration resource
+		deleteWebhookAssets = append(deleteWebhookAssets, ValidatingWebhookConfiguration)
+		applyWebhookAssets = getSubsetOfAssets(applyWebhookAssets, ValidatingWebhookConfiguration)
 	}
 
 	if !mutatingWebhookEnabled {
-		// Remove mutating resources
-		deleteAssets = append(deleteAssets, mutatingCRDs...)
-		deleteAssets = append(deleteAssets, MutatingWebhookConfiguration)
-		applyAssets = getSubsetOfAssets(applyAssets, mutatingCRDs...)
-		webhookAssets = getSubsetOfAssets(webhookAssets, MutatingWebhookConfiguration)
+		// Remove and apply mutating resources
+		deleteWebhookAssets = append(deleteWebhookAssets, MutatingWebhookConfiguration)
+		applyOrderedAssets = getSubsetOfAssets(applyOrderedAssets, MutatingCRDs...)
+		applyWebhookAssets = getSubsetOfAssets(applyWebhookAssets, MutatingWebhookConfiguration)
+		deleteCRDAssets = append(deleteCRDAssets, MutatingCRDs...)
 	}
 	return
 }
@@ -367,9 +410,14 @@ func (r *GatekeeperReconciler) crudResource(obj *unstructured.Unstructured, gate
 
 	logger := r.Log.WithValues("Gatekeeper resource", namespacedName)
 
-	err = ctrl.SetControllerReference(gatekeeper, obj, r.Scheme)
-	if err != nil {
-		return errors.Wrapf(err, "Unable to set controller reference for %s", namespacedName)
+	// Skip adding a controller reference for the namespace so that a deletion
+	// of the Gatekeeper CR does not also delete the namespace and everything
+	// within it.
+	if obj.GetKind() != util.NamespaceKind {
+		err = ctrl.SetControllerReference(gatekeeper, obj, r.Scheme)
+		if err != nil {
+			return errors.Wrapf(err, "Unable to set controller reference for %s", namespacedName)
+		}
 	}
 
 	err = r.Get(ctx, namespacedName, clusterObj)
@@ -410,7 +458,7 @@ func (r *GatekeeperReconciler) crudResource(obj *unstructured.Unstructured, gate
 }
 
 func (r *GatekeeperReconciler) isOpenShift() bool {
-	return util.IsOpenShift(r.PlatformName)
+	return r.PlatformInfo.IsOpenShift()
 }
 
 var commonSpecOverridesFn = []func(*unstructured.Unstructured, operatorv1alpha1.GatekeeperSpec) error{
@@ -419,6 +467,7 @@ var commonSpecOverridesFn = []func(*unstructured.Unstructured, operatorv1alpha1.
 	setPodAnnotations,
 	setTolerations,
 	containerOverrides,
+	setEnableMutation,
 }
 var commonContainerOverridesFn = []func(map[string]interface{}, operatorv1alpha1.GatekeeperSpec) error{
 	setImage,
@@ -461,11 +510,6 @@ func crOverrides(gatekeeper *operatorv1alpha1.Gatekeeper, asset string, obj *uns
 				return err
 			}
 		}
-		if mutatingWebhookEnabled(gatekeeper.Spec.MutatingWebhook) {
-			if err := setEnableMutation(obj); err != nil {
-				return err
-			}
-		}
 	// ValidatingWebhookConfiguration overrides
 	case ValidatingWebhookConfiguration:
 		if err := webhookConfigurationOverrides(obj, gatekeeper.Spec.Webhook, ValidationGatekeeperWebhook, true, controllerDeploymentPending); err != nil {
@@ -543,18 +587,26 @@ func webhookOverrides(obj *unstructured.Unstructured, webhook *operatorv1alpha1.
 		if err := setResources(obj, webhook.Resources); err != nil {
 			return err
 		}
+		if err := setDisabledBuiltins(obj, webhook.DisabledBuiltins); err != nil {
+			return err
+		}
 	}
 	return nil
 }
 
 func webhookConfigurationOverrides(obj *unstructured.Unstructured, webhook *operatorv1alpha1.WebhookConfig, webhookName string, updateFailurePolicy bool, controllerDeploymentPending bool) error {
+	// Set failure policy to ignore if deployment is still pending.
+	if controllerDeploymentPending {
+		ignore := admregv1.Ignore
+		failurePolicy := &ignore
+		if err := setFailurePolicy(obj, failurePolicy, webhookName); err != nil {
+			return err
+		}
+	}
+
 	if webhook != nil {
-		if updateFailurePolicy || controllerDeploymentPending {
+		if updateFailurePolicy && !controllerDeploymentPending {
 			failurePolicy := webhook.FailurePolicy
-			if controllerDeploymentPending {
-				ignore := admregv1.Ignore
-				failurePolicy = &ignore
-			}
 			if err := setFailurePolicy(obj, failurePolicy, webhookName); err != nil {
 				return err
 			}
@@ -663,21 +715,21 @@ func removeAnnotations(obj *unstructured.Unstructured) error {
 
 func setLogLevel(obj *unstructured.Unstructured, logLevel *operatorv1alpha1.LogLevelMode) error {
 	if logLevel != nil {
-		return setContainerArg(obj, managerContainer, LogLevelArg, string(*logLevel))
+		return setContainerArg(obj, managerContainer, LogLevelArg, string(*logLevel), false)
 	}
 	return nil
 }
 
 func setAuditInterval(obj *unstructured.Unstructured, auditInterval *metav1.Duration) error {
 	if auditInterval != nil {
-		return setContainerArg(obj, managerContainer, AuditIntervalArg, fmt.Sprint(auditInterval.Round(time.Second).Seconds()))
+		return setContainerArg(obj, managerContainer, AuditIntervalArg, fmt.Sprint(auditInterval.Round(time.Second).Seconds()), false)
 	}
 	return nil
 }
 
 func setConstraintViolationLimit(obj *unstructured.Unstructured, constraintViolationLimit *uint64) error {
 	if constraintViolationLimit != nil {
-		return setContainerArg(obj, managerContainer, ConstraintViolationLimitArg, strconv.FormatUint(*constraintViolationLimit, 10))
+		return setContainerArg(obj, managerContainer, ConstraintViolationLimitArg, strconv.FormatUint(*constraintViolationLimit, 10), false)
 	}
 	return nil
 }
@@ -688,14 +740,14 @@ func setAuditFromCache(obj *unstructured.Unstructured, auditFromCache *operatorv
 		if *auditFromCache == operatorv1alpha1.AuditFromCacheEnabled {
 			auditFromCacheValue = "true"
 		}
-		return setContainerArg(obj, managerContainer, AuditFromCacheArg, auditFromCacheValue)
+		return setContainerArg(obj, managerContainer, AuditFromCacheArg, auditFromCacheValue, false)
 	}
 	return nil
 }
 
 func setAuditChunkSize(obj *unstructured.Unstructured, auditChunkSize *uint64) error {
 	if auditChunkSize != nil {
-		return setContainerArg(obj, managerContainer, AuditChunkSizeArg, strconv.FormatUint(*auditChunkSize, 10))
+		return setContainerArg(obj, managerContainer, AuditChunkSizeArg, strconv.FormatUint(*auditChunkSize, 10), false)
 	}
 	return nil
 }
@@ -706,13 +758,33 @@ func setEmitEvents(obj *unstructured.Unstructured, argName string, emitEvents *o
 		if *emitEvents == operatorv1alpha1.EmitEventsEnabled {
 			emitArgValue = "true"
 		}
-		return setContainerArg(obj, managerContainer, argName, emitArgValue)
+		return setContainerArg(obj, managerContainer, argName, emitArgValue, false)
 	}
 	return nil
 }
 
-func setEnableMutation(obj *unstructured.Unstructured) error {
-	return setContainerArg(obj, managerContainer, EnableMutationArg, "true")
+func setDisabledBuiltins(obj *unstructured.Unstructured, disabledBuiltins []string) error {
+	for _, b := range disabledBuiltins {
+		if err := setContainerArg(obj, managerContainer, DisabledBuiltinArg, b, true); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func setEnableMutation(obj *unstructured.Unstructured, spec operatorv1alpha1.GatekeeperSpec) error {
+	if !mutatingWebhookEnabled(spec.MutatingWebhook) {
+		return nil
+	}
+
+	switch obj.GetName() {
+	case AuditDeploymentName:
+		return setContainerArg(obj, managerContainer, OperationArg, OperationMutationStatus, true)
+	case WebhookDeploymentName:
+		return setContainerArg(obj, managerContainer, EnableMutationArg, "true", false)
+	default:
+		return nil
+	}
 }
 
 func setWebhookConfigurationWithFn(obj *unstructured.Unstructured, webhookName string, webhookFn func(map[string]interface{}) error) error {
@@ -821,14 +893,18 @@ func setResources(obj *unstructured.Unstructured, resources *corev1.ResourceRequ
 }
 
 func setImage(container map[string]interface{}, spec operatorv1alpha1.GatekeeperSpec) error {
-	if spec.Image == nil {
-		return nil
-	}
-	if spec.Image.Image != nil {
-		if err := unstructured.SetNestedField(container, *spec.Image.Image, "image"); err != nil {
+	image := os.Getenv(GatekeeperImageEnvVar)
+	if image != "" {
+		if err := unstructured.SetNestedField(container, image, "image"); err != nil {
 			return errors.Wrapf(err, "Failed to set container image")
 		}
 	}
+	// else should only happen in dev/test environments, in which case use the
+	// default image in the Gatekeeper deployment manifests i.e. no overrides.
+
+	if spec.Image == nil {
+		return nil
+	}
 	if spec.Image.ImagePullPolicy != nil {
 		if err := unstructured.SetNestedField(container, string(*spec.Image.ImagePullPolicy), "imagePullPolicy"); err != nil {
 			return errors.Wrapf(err, "Failed to set container image pull policy")
@@ -858,7 +934,7 @@ func setContainerAttrWithFn(obj *unstructured.Unstructured, containerName string
 	return nil
 }
 
-func setContainerArg(obj *unstructured.Unstructured, containerName, argName string, argValue string) error {
+func setContainerArg(obj *unstructured.Unstructured, containerName, argName string, argValue string, isMultiArg bool) error {
 	return setContainerAttrWithFn(obj, containerName, func(container map[string]interface{}) error {
 		args, found, err := unstructured.NestedStringSlice(container, "args")
 		if !found || err != nil {
@@ -866,8 +942,8 @@ func setContainerArg(obj *unstructured.Unstructured, containerName, argName stri
 		}
 		exists := false
 		for i, arg := range args {
-			n, _ := util.FromArg(arg)
-			if n == argName {
+			n, v := util.FromArg(arg)
+			if n == argName && (!isMultiArg || isMultiArg && v == argValue) {
 				args[i] = util.ToArg(argName, argValue)
 				exists = true
 			}
@@ -916,7 +992,7 @@ func setControllerManagerExceptNamespace(obj *unstructured.Unstructured, asset,
 	if asset != WebhookFile {
 		return nil
 	}
-	return setContainerArg(obj, managerContainer, ExemptNamespaceArg, namespace)
+	return setContainerArg(obj, managerContainer, ExemptNamespaceArg, namespace, false)
 }
 
 func setRoleBindingSubjectNamespace(obj *unstructured.Unstructured, asset, namespace string) error {
